class hilight:
    """
        class hilight(string).color(highlight=True, bold=True)

        WARNING:
            This version no longer works as of Python v3.9. Use hilightV2 instead

        EXAMPLE:
            print (hilight("Hello World").red(bold=True))

            Or you could make an object:
                text = hilight("Bar")

                print ("Foo "+text.blue())

            To return the original string:
                print (text.string)
        
        COLORS:
            red
            green
            yellow
            blue
            purple
            teal
            white

        FUN FACTS:
            * This class is loosely based off the very first bit of python code I ever wrote. It was initially created while teaching myself python. 
            * This module was intentionally misspelled to shorten the keystrokes needed during use. 
    """
    def __init__(self, string):
        self.string = string


    def red(self, bold=False, highlight=False):
        """ hilight(string).red(bold=False, highlight=False) """
        attr = []

        if highlight:
            attr.append('41')
        else:
            attr.append('31')

        if bold:
            attr.append('1')

        return '\x1b[%sm%s\x1b[0m' % (';'.join(attr), self.string)


    def green(self, bold=False, highlight=False):
        """ hilight(string).green(bold=False, highlight=False) """
        attr = []

        if highlight:
            attr.append('42')
        else:
            attr.append('32')

        if bold:
            attr.append('1')

        return '\x1b[%sm%s\x1b[0m' % (';'.join(attr), self.string)


    def yellow(self, bold=False, highlight=False):
        """ hilight(string).yellow(bold=False, highlight=False) """
        attr = []

        if highlight:
            attr.append('43')
        else:
            attr.append('33')

        if bold:
            attr.append('1')

        return '\x1b[%sm%s\x1b[0m' % (';'.join(attr), self.string)


    def blue(self, bold=False, highlight=False):
        """ hilight(string).blue(bold=False, highlight=False) """
        attr = []

        if highlight:
            attr.append('44')
        else:
            attr.append('34')

        if bold:
            attr.append('1')

        return '\x1b[%sm%s\x1b[0m' % (';'.join(attr), self.string)


    def purple(self, bold=False, highlight=False):
        """ hilight(string).purple(bold=False, highlight=False) """
        attr = []

        if highlight:
            attr.append('45')
        else:
            attr.append('35')

        if bold:
            attr.append('1')

        return '\x1b[%sm%s\x1b[0m' % (';'.join(attr), self.string)


    def teal(self, bold=False, highlight=False):
        """ hilight(string).teal(bold=False, highlight=False) """
        attr = []

        if highlight:
            attr.append('46')
        else:
            attr.append('36')

        if bold:
            attr.append('1')

        return '\x1b[%sm%s\x1b[0m' % (';'.join(attr), self.string)


    def white(self, bold=False, highlight=False):
        """ hilight(string).white(bold=False, highlight=False) """
        attr = []

        if highlight:
            attr.append('47')
        else:
            attr.append('37')

        if bold:
            attr.append('1')

        return '\x1b[%sm%s\x1b[0m' % (';'.join(attr), self.string)

class hilightV2:
    """
    A class for creating colored and formatted terminal output.

    Example:
        print(hilightV2("Hello World").red(bold=True))

        Or you could make an object:
            text = hilightV2("Bar")
            print("Foo " + text.blue())

        To return the original string:
            print(text.get_text())

    Colors:
        - red
        - green
        - yellow
        - blue
        - purple
        - teal
        - white

    Fun Facts:
        - This class is loosely based on Python code I wrote while learning the language.
        - The module name is intentionally misspelled to shorten keystrokes.
    """

    def __init__(self, text):
        """
        Initialize the hilightV2 object with the input text.
        """
        self._text = text

    def _apply_format(self, color_code, bold):
        """
        Apply formatting using ANSI escape codes.
        """
        attr = [color_code]

        if bold:
            attr.append('1')

        return '\x1b[%sm%s\x1b[0m' % (';'.join(attr), self._text)

    def red(self, bold=False):
        """Format the text with red color."""
        return self._apply_format('31', bold)

    def green(self, bold=False):
        """Format the text with green color."""
        return self._apply_format('32', bold)

    def yellow(self, bold=False):
        """Format the text with yellow color."""
        return self._apply_format('33', bold)

    def blue(self, bold=False):
        """Format the text with blue color."""
        return self._apply_format('34', bold)

    def purple(self, bold=False):
        """Format the text with purple color."""
        return self._apply_format('35', bold)

    def teal(self, bold=False):
        """Format the text with teal color."""
        return self._apply_format('36', bold)

    def white(self, bold=False):
        """Format the text with white color."""
        return self._apply_format('37', bold)

    def get_text(self):
        """Return the original text."""
        return self._text



class HTMLgen:
    """
        Basic HTML generator
        02/09/2019

        class HTMLgen(head=False, tail=False, lang="en", docType="html")

        FUNCTIONS:
            title(self, title, scripts=None, css=None)
            body.add(content)
            tag(tag, content=False, close=True, cssclass=None)
            image(src, alt=None, srcset=None, height=None, width=None, style=None, cssclass=None)
            br() <-- Returns a </ br> tag
            div(cssclass) <-- Not yet implemented 
            table() -- Class Object
                table.add_row()
                table.produce_table()

        EXAMPLE:
            page = HTMLgen(True, True)
            page.title("This is the page Title", scripts="foo.js bar.js", css="styles.css nav.css")
            page.body.add(page.image("images/frontpage.jpg", width="100%"))
            page.body.add(page.tag("h1", "This is a header line"))
            page.body.add("This is another line")

            ## Creating a table
            test_table = HTMLgen.table()
            test_table.add_row(["column data 1", "column data 2", "column data 3"])
            test_table.add_row(["column data 4", "column data 5", "column data 6"])

            ## And finally we add the table to the rest of the page
            page.body.add(test_table.produce_table())


            page.return_html()

            ```
                <!DOCTYPE html>
                <html lang="en>
                    <head>
                        <title>This is the page Title</title>
                        <link rel="stylesheet" href="styles.css">
                        <link rel="stylesheet" href="nav.css">
                        <script src="foo.js"></script>
                        <script src="bar.js"></script>
                    </head>
                <body>
                    <img src="images/frontpage.jpg" width="100%">
                    <h1>This is a header line</h1>
                    This is another line
                   	<table>
                        <tr>
                            <td>column data 1</td><td>column data 2</td><td>column data 3</td>
                        </tr>
                        <tr>
                            <td>column data 4</td><td>column data 5</td><td>column data 6</td>
                        </tr>

                    </table>

                </body>
                </html>
            ```

    """

    def __init__(self, head=False, tail=False, lang="en", docType="html"):
        self.head = head
        self.tail = tail
        self.page = {}
        self.body = self.createBody()

        if self.head:
            self.page["head"] = "<!DOCTYPE "+docType+">\n"
            self.page["head"] = self.page["head"]+"<html lang=\""+lang+"\">\n"

        if self.tail:
            self.page["tail"] = '</html>'


    def __exit__(self, exception_type, exception_value, traceback):
        self.return_html(self)


    def title(self, title, scripts=None, css=None):
        output = "\t<head>\n"
        output = output + "\t\t<title>"+title+"</title>\n"
        if css:
            if " " in css:
                css = css.split(' ')
                for i in css:
                    output = output + "\t\t<link rel=\"stylesheet\" href=\""+i+"\">\n"
            else:
                output = output + "\t\t<link rel=\"stylesheet\" href=\""+css+"\">\n"
        if scripts:
            if " " in scripts:
                scripts = scripts.split(' ')
                for script in scripts:
                    output = output + "\t\t<script src=\""+script+"\"></script>\n"
            else:
                output = output + "\t\t<script src=\""+scripts+"\"></script>\n"
            
        output = output + "\t</head>\n"

        self.page["header"] = output


    def tag(self, tag, content=False, close=True, cssclass=None):
        output = "<"+tag
        if cssclass:
            output = output+" class=\""+cssclass+'">'
        else:
            output = output+">"

        if content:
            output = output+content
        if close:
            output = output+"</"+tag+">"

        return output


    def br(self):
        return "</ br>"


    def createBody(self):
        return HTMLgen.bodycontent(self)


    class bodycontent:
        def __init__(self, body):
            self.body = body
            self.content = []

        def add(self, content):
            n = 25
            strings = content.split(' ')
            if len(strings) > n:
                count = 0
                new_content_count = 0
                new_content = []
                content = ""
                for string in strings:
                    if count <= n:
                        count += 1
                    else: 
                        count = 0
                        new_content_count += 1

                    try:
                        new_content[new_content_count] = new_content[new_content_count] + " " + string
                    except IndexError:
                        new_content.append(string)

                for string in new_content:
                    content = content + "\t" + string + "\n"
            
            else:
                content = "\t"+content


            self.content.append(content)


    class table:
        def __init__(self, border=None, cellpadding=None):
            self.rows = []
            self.border = border
            self.cellpadding = cellpadding

        def add_row(self, data):
            """
                Row data should be a list of data
                EXAMPLE:
                    table.add_row(["col 1", "col 2"])

                    To add colspan:
                    table.add_row(["colspan:2_col3"])

                    to add colspan and align text:
                    table.add_row(["colspan:2 align=center_col3])
            """
            #data = data + "\n"
            self.rows.append(data)

        def produce_table(self):
            table_output = "<table"
            if self.border:
                table_output = table_output+" border={}".format(self.border)

            if self.cellpadding:
                table_output = table_output+" cellpadding={}".format(self.cellpadding)

            table_output = table_output+">\n"

            for row in self.rows:
                table_output = table_output+"\t\t<tr>\n\t\t\t"
                for column_data in row:
                    colspan_value = None
                    if "colspan:" in column_data:
                        colspan_value = str(column_data).split('_')[0]
                        column_data = str(column_data).split('_')[1]
                        colspan_value = colspan_value.split(':')[1]

                    if colspan_value:
                        table_output = table_output+"<td colspan={}>{}</td>".format(colspan_value, column_data)
                    else:
                        table_output = table_output+"<td>"+str(column_data)+"</td>"
                
                table_output = table_output+"\n"
                table_output = table_output+"\t\t</tr>\n"

            table_output = table_output+"\n\t</table>"

            return table_output

    
    def image(self, src, alt=None, srcset=None, height=None, width=None, style=None, cssclass=None):
        if src:
            image = '<img src="'+src+'"'
        else:
            return False

        if alt:
            image = image + ' alt="'+alt+'"'

        if srcset:
            image = image + ' srcset="'+srcset+'"'

        if height:
            image = image + ' height="'+height+'"'

        if width:
            image = image + ' width="'+width+'"'

        if style:
            image = image + ' style="'+style+'"'

        if cssclass:
            image = image + " class=\""+cssclass+'"'

        image = image + ">"

        return str(image)


    def return_html(self):
        output = ""
        if "head" in self.page:
            output = output + self.page["head"]
        if "header" in self.page:
            output = output + self.page["header"]
        if self.body.content:
            output = output + "<body>\n"
            for i in self.body.content:
                output = output + i + "\n"
            output = output + "</body>\n"
        if "tail" in self.page:
            output = output + self.page["tail"]

        return output


import sys

def progress_bar(progress, barLength=50, text_field="Progress"):
    """
    Displays or updates a console progress bar.

    Accepts a float between 0 and 1. Any int will be converted to a float.
    A value under 0 represents a 'halt'.
    A value at 1 or bigger represents 100%.

    Parameters:
    - progress: A float between 0 and 1.
    - barLength: Length of the progress bar in characters.
    - text_field: Text to display before the progress bar.
    """
    status = " \r"
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "ERROR: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done...\r\n"
    block = int(round(barLength * progress))
    text = "\r{3}: [{0}] {1}%  {2}".format("#" * block + "-" * (barLength - block), round(progress * 100), status, text_field)
    print(text, end='\r')

def colored_progress_bar(progress, barLength=50, text_field="Progress"):
    """
    Displays or updates a console progress bar with color changing when it's greater than 99%.

    Parameters:
    - progress: A float between 0 and 1.
    - barLength: Length of the progress bar in characters.
    - text_field: Text to display before the progress bar.
    """
    def set_color_code(color):
        colors = {
            'green': '\033[92m',
            'red': '\033[91m',
            'yellow': '\033[93m',
            'white': '\033[97m',
            'purple': '\033[95m',
            'cyan': '\033[96m',
        }
        return colors.get(color, '\033[0m')

    if progress >= 0.94:
        color = 'green'
    elif progress >= 0.4:
        color = 'yellow'
    else:
        color = 'red'
    color_code = set_color_code(color)

    status = " \r"
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "ERROR: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done...\r\n"
    block = int(round(barLength * progress))
    text = f"{text_field}: {color_code}[{'#'*block + '-'*(barLength-block)}] {round(progress*100)}% {status}\033[0m"
    sys.stdout.write(text)
    sys.stdout.flush()